import datetime
import json
from urllib.parse import urljoin

import inflection
import pytz
import requests

from toss.entity import Payment
from toss.exc import NotAutoExecutable
from toss.response import (ApprovedResult, APIResponse, APIError,
                           PurchaseResult, CancelledResult, RefundedResult)
from toss.validator import validate_order_number


class TossPayClient:

    base_url = 'https://pay.toss.im/'
    api_version = 'v1'
    production_api_key = 'sk_test_apikey1234567890'
    development_api_key = 'sk_test_apikey1234567890'
    sandbox_key = 'sk_test_apikey1234567890'
    development = True

    def __init__(self,
                 production_api_key: str=None,
                 development_api_key: str=None,
                 development: bool=True):
        self.production_api_key = production_api_key or self.sandbox_key
        self.development_api_key = development_api_key or self.sandbox_key
        self.development = development

    @property
    def api_key(self):
        if self.development:
            return self.development_api_key
        return self.production_api_key

    def build_url(self, uri: str) -> str:
        return urljoin(self.base_url,
                       urljoin('/api/{}/'.format(self.api_version), uri))

    def request(self, method: str, uri: str, params: dict) -> APIResponse:
        params['apiKey'] = self.api_key

        filtered = {}
        for k, v in params.items():
            if v is not None:
                filtered[k] = v

        result = getattr(requests, method)(self.build_url(uri), data=filtered)

        try:
            jsonized = result.json()
        except json.decoder.JSONDecodeError:
            return APIError('unsupported api response',
                            response=result.text)

        return APIResponse(**jsonized)

    def purchase(self,
                 order_no: str,
                 amount: int,
                 product_desc: str,
                 ret_url: str,
                 cash_receipt: bool,
                 amount_tax_free: int=0,
                 auto_execute: bool=False,
                 amount_taxable: int=0,
                 amount_vat: int=0,
                 amount_service_fee: int=0,
                 expired_time: datetime=datetime.timedelta(minutes=15),
                 result_callback: str='',
                 escrow: bool=False,
                 checkout_type: str='web',
                 ars_auth_skippable: str='Y',
                 user_phone: str='',
                 partner_id: str='',
                 metadata: str='',
                 ret_cancel_url: str='') -> (PurchaseResult or APIResponse):

        if not result_callback and auto_execute:
            raise NotAutoExecutable
        if expired_time > datetime.timedelta(hours=1):
            raise ValueError("`expired_time` exceeds 1 hour")

        validate_order_number(order_no)

        basic_params = {
            "orderNo": order_no,
            "amount": amount,
            "productDesc": product_desc,
            "autoExecute": auto_execute or None,
            "retUrl": ret_url,
            "amountTaxable": amount_taxable or None,
            "amountTaxFree": amount_tax_free,
            "amountVat": amount_vat or None,
            "amountServiceFee": amount_service_fee or None,
            "expiredTime": (
                datetime.datetime.now(
                    tz=pytz.timezone('Asia/Seoul')) + expired_time
            ).strftime("%Y-%m-%d %H:%M:%S"),
            "resultCallback": result_callback or None,
            "escrow": escrow or None,
            "cashReceipt": cash_receipt or None,
            "checkoutType": checkout_type or None,
            "arsAuthSkippable": ars_auth_skippable or None,
            "userPhone": user_phone or None,
            "partnerId": partner_id or None,
            "metadata": metadata or None,
            "retCancelUrl": ret_cancel_url or None
        }

        result = self.request('post', 'payments', basic_params)

        if result.data['code'] == -1:
            return result

        return PurchaseResult(pay_token=result.data["payToken"],
                              purchase_url=result.data["checkoutPage"],
                              client=self,
                              code=result.data["code"])

    def get_payment(self, pay_token: str=None, order_no: str=None) -> Payment:
        if not pay_token and order_no:
            raise ValueError("`pay_token` or `order_no` is necessary")

        result = self.request('post', 'status', {'payToken': pay_token})

        params = dict((inflection.underscore(k), v)
                      for k, v in result.data.items())

        return Payment(**params)

    def approve(self,
                pay_token: str,
                amount: int=None,
                order_no: int=None) -> (APIResponse or ApprovedResult):

        params = {'payToken': pay_token}

        if amount:
            params['amount'] = amount
        if order_no:
            params['orderNo'] = order_no

        result = self.request('post', 'execute', params)

        if result.data['code'] == -1:
            return result

        return ApprovedResult(code=result.data['code'],
                              approved_at=result.data['approvalTime'])

    def cancel(self, pay_token: str, reason: str) -> APIResponse:

        params = {'payToken': pay_token, 'reason': reason}

        result = self.request('post', 'cancel', params)

        if result.data['code'] == -1:
            return result

        return CancelledResult(code=result.data['code'])

    def refund(self,
               pay_token: str,
               amount: int,
               amount_tax_free: int,
               refund_no: str=None,
               reason: str=None,
               amount_taxable: int=None,
               amount_vat: int=None,
               amount_service_fee: int=None) -> APIResponse:

        params = {'payToken': pay_token, 'refundNo': refund_no,
                  'reason': reason, 'amount': amount,
                  'amountTaxable': amount_taxable,
                  'amountTaxFree': amount_tax_free, 'amountVat': amount_vat,
                  'amountServiceFee': amount_service_fee}

        result = self.request('post', 'refunds', params)

        if result.data['code'] == -1:
            return result

        return RefundedResult(code=result.data['code'],
                              refund_no=result.data['refundNo'],
                              approved_at=result.data['approvalTime'])



from datetime import timedelta
from uuid import uuid4

import pytest
import requests_mock

import toss
from toss.entity import Payment
from toss.response import (APIError, PurchaseResult, ApprovedResult,
                           CancelledResult, RefundedResult)


def test_purchase():
    c = toss.TossPayClient(development=True)
    order_id = str(uuid4())

    with pytest.raises(ValueError):
        c.purchase(order_id + '&', 40000, 'test', '', True)

    with pytest.raises(ValueError):
        c.purchase(order_id, 40000, 'test', '', True,
                   expired_time=timedelta(hours=1, minutes=1))

    with pytest.raises(toss.exc.NotAutoExecutable):
        c.purchase(order_id, 40000, 'test', '', True,
                   auto_execute=True)

    result = c.purchase(order_id, 40000, 'test', '', True)

    assert result.pay_token is not None
    assert result.purchase_url is not None

    order_id = str(uuid4())
    result = c.purchase(order_id, 40000, 'test', '', True,
                        auto_execute=True, result_callback='test')

    assert result.pay_token is not None
    assert result.purchase_url is not None
    assert isinstance(result, PurchaseResult)

    result = c.purchase(order_id, 0, 'test', '', True,
                        auto_execute=True, result_callback='test')
    assert isinstance(result, APIError)
    assert result.data["status"] == 200
    assert result.data["code"] == -1
    assert result.msg == '요청한 값이 부족하거나 올바르지 않습니다. amount는 0보다 커야 합니다.'
    assert result.data["errorCode"] == 'COMMON_INVALID_PARAMETER'


def test_get_payment():
    pay_token = "N4GOTJB5eR3Tnx8kJeVp90"

    c = toss.TossPayClient(development=True)
    result = c.get_payment(pay_token)

    assert isinstance(result, Payment)


def test_purchase_result():

    c = toss.TossPayClient(development=True)
    order_id = str(uuid4())

    purchase_result = c.purchase(order_id, 40000, 'test', '', True)

    payment = c.get_payment(purchase_result.pay_token)

    assert isinstance(payment, Payment)
    assert payment.amount == 40000
    assert payment.product_desc == 'test'
    assert payment.pay_status == 'PAY_STANDBY'


def test_confirm_purchase():

    c = toss.TossPayClient(development=True)
    order_id = str(uuid4())
    purchase_result = c.purchase(order_id, 40000, 'test', '', True)
    payment = c.get_payment(purchase_result.pay_token)

    approved_result = c.approve(payment.pay_token)

    assert isinstance(approved_result, APIError)
    assert approved_result.msg == '사용자 정보가 존재하지 않습니다.'

    result = c.purchase(order_id, 40000, 'test', '', True)
    payment = c.get_payment(result.pay_token)
    token = payment.pay_token

    with requests_mock.Mocker() as m:
        # NOTE: toss user-side auth 가 자동화될 수가 없어 mocking 으로 우회
        m.post('https://pay.toss.im/api/v1/execute',
               text='{"code":0,"approvalTime":"2016-11-16 13:59:59"}')
        approved_result = c.approve(token)

    assert isinstance(approved_result, ApprovedResult)


def test_cancel_purchase():

    c = toss.TossPayClient(development=True)
    order_id = str(uuid4())
    purchase_result = c.purchase(order_id, 40000, 'test', '', True)
    payment = c.get_payment(purchase_result.pay_token)

    cancelled_result = c.cancel(payment.pay_token, 'test cancel')

    assert isinstance(cancelled_result, CancelledResult)

    order_id = str(uuid4())
    pr = c.purchase(order_id, 40000, 'test', '', True, auto_execute=True,
                    result_callback='test')
    payment = c.get_payment(pr.pay_token)
    token = payment.pay_token
    cancelled_result = c.cancel(token, 'test cancel')

    assert isinstance(cancelled_result, CancelledResult)

    order_id = str(uuid4())
    result = c.purchase(order_id, 40000, 'test', '', True)
    payment = c.get_payment(result.pay_token)
    token = payment.pay_token

    with requests_mock.Mocker() as m:
        # NOTE: toss user-side auth 가 자동화될 수가 없어 mocking 으로 우회
        m.post('https://pay.toss.im/api/v1/execute',
               text='{"code":0,"approvalTime":"2016-11-16 13:59:59"}')
        c.approve(token)
        # NOTE: 승인이 되어 결제가 완료된 시점에서는 취소가 불가능함
        m.post('https://pay.toss.im/api/v1/cancel',
               text='{"code": -1, "errorCode": "CANCEL_IMPOSSIBLE_STATUS", '
                    '"status": 200, "msg": "취소가 불가능한 상태입니다."}')
        cancelled_result = c.cancel(token, 'test')
        assert isinstance(cancelled_result, APIError)
        assert cancelled_result.data['errorCode'] == 'CANCEL_IMPOSSIBLE_STATUS'


def test_cancel_refund():

    c = toss.TossPayClient(development=True)
    order_id = str(uuid4())
    purchase_result = c.purchase(order_id, 40000, 'test', '', True)
    payment = c.get_payment(purchase_result.pay_token)
    refund_result = c.refund(payment.pay_token, 40000, 0)
    assert isinstance(refund_result, APIError)

    # NOTE: toss user-side auth 가 자동화될 수가 없어 mocking 으로 우회
    # NOTE: 이후는 승인되었다고 가정

    token = payment.pay_token

    with requests_mock.Mocker() as m:
        m.post('https://pay.toss.im/api/v1/refunds',
               text='{"code": 0, "refundNo": "b764f9ec-a113-401d-9db8-'
                    '532a36c0986b", "approvalTime": "2018-08-27 13:25:11"}')
        refunded_result = c.refund(token, 40000, 0)
        assert isinstance(refunded_result, RefundedResult)



import re


def validate_order_number(order_no: str) -> None:
    if len(order_no) > 50:
        raise ValueError("`order_no` length should be lesser than 50")
    if not re.match('^(\w|[_\-:.^@\'])+$', order_no):
        raise ValueError("`order_no` is not valid pattern")



from toss.entity import Payment


class APIResponse:

    def __new__(cls, *args, **kwargs):
        if kwargs["code"] == -1:
            return super(cls, APIError).__new__(APIError)
        return super(APIResponse, cls).__new__(cls)

    def __init__(self, **kwargs):
        self.data = kwargs


class APIError(APIResponse):

    def __init__(self, msg: str, **kwargs):
        self.msg = msg
        super(APIError, self).__init__(**kwargs)


class PurchaseResult(APIResponse):
    def __init__(self, pay_token: str, purchase_url: str, *args, **kwargs):
        self.pay_token = pay_token
        self.purchase_url = purchase_url
        super(PurchaseResult, self).__init__(**kwargs)


class ApprovedResult(APIResponse):

    def __init__(self, code: str, approved_at: str, **kwargs):
        self.code = code
        self.approved_at = approved_at
        super(ApprovedResult, self).__init__(**kwargs)


class CancelledResult(APIResponse):

    pass


class RefundedResult(APIResponse):

    def __init__(self, code: str, refund_no: str, approved_at: str, **kwargs):
        self.code = code
        self.refund_no = refund_no
        self.approved_at = approved_at
        super(RefundedResult, self).__init__(**kwargs)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="toss",
    version="0.0.1",
    author="Sungyong Kang",
    author_email="hiye@styleshare.kr",
    description="Toss payment client library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/styleshare/toss",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Public Domain",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'inflection==0.3.1',
        'pytz==2018.5',
        'requests==2.19.1'
    ]
)





class BaseEntity:

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class Payment(BaseEntity):
    def __init__(self, pay_token: str, pay_status: str, order_no: str,
                 amount: int,
                 amount_taxable: int, amount_tax_free: int, amount_vat: int,
                 amount_service_fee: int, time_created:str,
                 time_pay_complete: str,
                 time_pay_cancel: str, product_desc: str, has_owner: bool,
                 available_actions: list,
                 refunds: list, metadata: str, **kwargs):
        self.pay_token = pay_token
        self.pay_status = pay_status
        self.order_no = order_no
        self.amount = amount
        self.amount_taxable = amount_taxable
        self.amount_tax_free = amount_tax_free
        self.amount_vat = amount_vat
        self.amount_service_fee = amount_service_fee
        self.time_created = time_created
        self.time_pay_complete = time_pay_complete
        self.time_pay_cancel = time_pay_cancel
        self.product_desc = product_desc
        self.has_owner = has_owner
        self.available_actions = available_actions
        self.refunds = refunds
        self.metadata = metadata
        super(Payment, self).__init__(**kwargs)
